------------------------------------------------------------------------------------------
-- psp-httpd, created by Elton Muuga / Elxx.net
------------------------------------------------------------------------------------------
-- srv.lua - Main server functions and main server thread
------------------------------------------------------------------------------------------
-- Note that this app uses a customized version of the LuaPlayer source, thus it will not 
-- work using an official release of LuaPlayer. You must use the included EBOOT.PBP.
------------------------------------------------------------------------------------------

appversion = "0.6"

-- Kill any existing WLAN connections, to prevent bind errors
-- This may cause problems with the 2.0 Eboot Loader, so you can comment it out if you want

--Wlan.term()

-- Start up the WLAN stuff
Wlan.init()

-- Set buffers and colors for the graphicsPrint function
white = Color.new(255, 255, 255)
blue = Color.new(58,174,255)
red = Color.new(255,0,0)
green = Color.new(0,255,0)
orange = Color.new(255,165,0)
bgblue = Color.new(0,90,173)
black = Color.new(0,0,0)
offscreen = Image.createEmpty(480, 272)

--Create timers
idleTicker = Timer.new()
uptimeCounter = Timer.new()

--
-- UTILITY FUNCTIONS
--

--  Print text on the screen, console-format - taken from WLAN example, with some
--  modifications for text overflow and colour codes.
function graphicsPrint(text)
	col = white
	if y > 272 then
		offscreen:clear(Color.new(0, 0, 0))
		y = 25
	end
	showStats()
	for i = 1, string.len(text) do
		char = string.sub(text, i, i)
		if char == "\n" then
			y = y + 10
			x = 5
		elseif char ~= "\r" then
			if char == "\1" then
				col = white
			elseif char == "\2" then
				col = blue
			elseif char == "\3" then
				col = red
			elseif char == "\4" then
				col = green
			elseif char == "\5" then
				col = orange
			else
				offscreen:print(x, y, char, col)
				x = x + 8
				if x > 470 then
					y = y + 8
					x = 5
				end
			end
		end
	end
	screen:blit(0, 0, offscreen)
	screen.waitVblankStart()
	screen.flip()
end

function showStats() -- Display blue bar on top of screen with bandwith and IP
	offscreen:fillRect(0, 0, 480, 15, bgblue)
	offscreen:print(5,5,"Out: ".. friendlySize(totalOutput),white)
	offscreen:print(120,5,"In: ".. friendlySize(totalInput),white)

	offscreen:print(220,5,"Up: " .. friendlyTime(uptimeCounter:time()),white)
	
	ip = Wlan.getIPAddress()
	if ip then
		offscreen:print(480 - (string.len(ip) * 8) - (8 * string.len(options.server_port)) - 8,5,ip .. ":" .. options.server_port,white)
	end
end

function urldecode(str) -- Decode special chars in URL's such as %20
	str = string.gsub (str, "+", " ") 
	str = string.gsub (str, "%%(%x%x)", function(h) return string.char(tonumber(h,16)) end) 
	str = string.gsub (str, "\r\n", "\n")
	return str
end

function strsplit(delimiter, text) -- Split a string into an array, going by a defined delimiter
	local list = {}
	local pos = 1
	if string.find("", delimiter, 1) then -- this would result in endless loops
		error("Header data error")
	end
	while 1 do
		local first, last = string.find(text, delimiter, pos)
		if first then -- found?
			table.insert(list, string.sub(text, pos, first-1))
			pos = last+1
		else
			table.insert(list, string.sub(text, pos))
			break
		end
	end
	return list
end

function fileSize(file)
	if file then
		local current = file:seek()      -- get current position
		local size = file:seek("end")    -- get file size
		file:seek("set", current)        -- restore position
		return size
	end
end

function friendlySize(size)
	if size >= 1048576 then				-- 1,048,576b = 1 MB
		return string.format("%.1f",(size / 1048576)) .. " MB"
	elseif size >= 1024 then				-- 1024b = 1 kB
		return string.format("%.0f",(size / 1024)) .. " kB"
	else 
		return size .. " b"
	end
end	

function friendlyTime(t)
	if t > 1000 * 60 * 60 then
		hours = math.floor(t / 1000 / 60 / 60)
	else hours = 0 end
	if t > 1000 * 60 then
		minutes = math.floor((t / 1000 / 60) - (60 * hours))
	else minutes = 0 end
	if t > 1000 then
		seconds = math.floor((t / 1000) - (60 * 60 * hours) - (60 * minutes))
	else seconds = 0 end

	return hours .. ":" .. string.format("%02d",minutes) .. "." .. string.format("%02d",seconds)
end

function string.endsWith(String,End)
	return End=='' or string.sub(String,-string.len(End))==End
end

function fileExists(file)
	t, msg = io.open(file, "r")
     if not t then return false end
	return true
end

--
-- SERVER FUNCTIONS
--

function start_server(port)

	-- Print initialization strings
	graphicsPrint("psp-httpd beta " .. appversion .. "\n")
	graphicsPrint("by Elxx - http://elxx.net\n")
	graphicsPrint("Uses portions of code from Steve Kemp\n")
	graphicsPrint("Thanks to modsyn for the network profile menu\n")
	graphicsPrint("+ Art and Xufar for testing and ideas\n\n")
	graphicsPrint("Data chunk size for file transfer is " .. options.chunksize .. " bytes.\n\n")

	running = 1;

	--  Create a server socket to listen on the port
	if socketCreated == 0 then
		serverSocket = Socket.createServerSocket( port )
		socketCreated = 1
	end

	--   Print some status messages.
	graphicsPrint("\4Server socket started and listening.\n\n")

	--  Loop accepting requests.
	while ( running == 1 ) do
		status = processConnection( serverSocket )
		if (status == "stop") or (status == "restart") then 
			running = 0 
			break
		end
	end

	-- Server stopped, close the socket.
	-- serverSocket:close()

	return status
end

function processConnection( serverSocket ) 
    --  Accept a new connection

	found   = 0;  -- Found the end of the HTTP headers?
	chunk   = 0;  -- Count of data read from client socket
	size    = 0;  -- Total size of incoming request.
	code    = 0;  -- Status code we send to the client
	request = ""; -- Request body read from client.

	_get = {}
	_post = {}

	length_defined = 0
	
	if options.auto_restart > 0 then
		idleTicker:reset()
		idleTicker:start()
	end

	--  Read in a response from the client, terminating at the first
	-- '\r\n\r\n' line - this is the end of the HTTP header request.

	while ( found == 0 ) do
		client = serverSocket:accept()
		if Controls.read():start() then return "stop" end
		if Controls.read():select() then return "restart" end
		if Controls.read():triangle() then 
			screen:save("screenshot.png")
			graphicsPrint("Screenshot saved.\n")
			System.sleep(2000)
		end
		System.powerTick()
		if options.auto_restart > 0 then
			if (idleTicker:time() / 1000) > options.auto_restart then 
				if options.autorestart_method == 1 then
					return "restart"
				else
					return "stop"
				end
			end
		end
		if client then
			graphicsPrint("Client found. Receiving request...\n")
			if options.auto_restart > 0 then idleTicker:reset() end
			while (found == 0) do
				data = client:advRecv()
				size    = size + string.len(data)
				request = request .. data
				_, _, type = string.find(request, "Content%-Type: (.-)\r\n")
				if type == "multipart/form-data" then
					graphicsPrint("\5found multipart form data\n")
					_, _, clen = string.find(request, "Content%-Length: ([^\r\n]+)")
					if clen then
						length_defined = 1
						hd_pos,hd_len = string.find( request, '\r\n\r\n' )
						if ( size >= tonumber(hd_len) ) then
							graphicsPrint("\5Received \4" .. size .. "\5 bytes\n")
							found = 1
						end
					end
				elseif type == "application/x-www-form-urlencoded" then
					found = 1
				else
					position,len = string.find( request, '\r\n\r\n' )
					if ( position ~= nil ) and ( length_defined == 0 )  then 
						found = 1 
					end
				end
				
				if Controls.read():start() then return "stop" end
				if Controls.read():select() then return "restart" end
			end
		end
    end

    totalInput = totalInput + size

    -- If we got this far without a client, that means something might've gone wrong...

	if not client then 
		return "restart"
	end

	--  We now have a complete HTTP request of 'size' bytes long stored in 'request'.
    
	-- Find the requested path.
	_, _, method, path, major, minor  = string.find(request, "([A-Z]+) (.+) HTTP/(%d).(%d)")

	if method == "POST" then
		--graphicsPrint(request)
		_, _, type = string.find(request, "Content%-Type: (.-)\r\n")
		if type == "application/x-www-form-urlencoded" then
			_, _, posts = string.find( request, "\r\n\r\n(.+)" )
			_post = get_post_vars( posts )
		else
			_, _, boundary = string.find(request, "boundary=([^\r\n]+)")
			_, _, posts = string.find( request, "\r\n\r\n(.+)" )
			_, _, clen = string.find(request, "Content%-Length: ([^\r\n]+)")
			graphicsPrint("\5Content-Length: \4" .. clen .. "\n")
			_post = get_multipart_vars(posts,boundary)
		end
	end

	-- Decode the requested path.

	path = urldecode( path )

	_get = get_url_vars( path )

	spl = strsplit("?", path)
	path = spl[1]
	
	-- Show the requested path on the PSP.
	graphicsPrint( "\2" .. path .. "\n")

	-- Find the user agent and referrer in case we need to log this someday.
	-- Virtual hosts are not yet supported.

	_, _, agent   = string.find(request, "Agent: ([^\r\n]+)")
	_, _, referer = string.find(request, "Referer: ([^\r\n]+)")
	-- _, _, host    = string.find(request, "Host: ([^:\r\n]+)")

	-- If there was no "Host:" header then use "default".
	if ( host == nil ) then
		host = "default"
	end

	size, code = handleRequest( path, client )

	-- if ( agent == nil )   then agent   = "-" end
	-- if ( referer == nil ) then referer = "-" end
	-- if ( code == nill )   then code    = 0 ; end

	if ( running == 0 ) then
		graphicsPrint( "\3Server terminated." )
	end

    --  Close the client connection.
    client:close();
end

function handleDirectory( client, file, path )

	h = io.open("Templates/head.tpl", "r")
	head = h:read("*a")
	h:close()
	tpl_output = "HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"
	tpl_output = tpl_output .. head

	files = System.listDirectory(file)
	if files then
		for index, file in files do
			if file.directory and file.name ~= "." then
				tpl_output = tpl_output .. "<p><a href=\"" .. file.name .. "/" .. "\">" .. file.name .. "</a></p>"
			end
		end
		for index, file in files do
			if not file.directory then
				tpl_output = tpl_output .. "<p><a href=\"" .. file.name .. "\">" .. file.name .. "</a> (" .. friendlySize(file.size) .. ")</p>"
			end
		end
	end

	h = io.open("Templates/footer.tpl", "r")
	footer = h:read("*a")
	h:close()

	tpl_output = tpl_output .. footer

	tpl_output = string.gsub(tpl_output, "{browsing_dir}", file)
	
	o = ""
	for word in string.gfind(file, "[.+^/]") do o = o .. "../" end

	tpl_output = string.gsub(tpl_output, "{uploader_location}", o .. "LHP/upload.lhp?loc=" .. file)

	send_data(tpl_output, client)
	graphicsPrint("\n")
end

function handleRequest( path, client )
	-- Local file
	file = path

	--  File must be on the Memory Stick.
	file = "ms0:" .. options.root_dir .. file

	-- If it is a directory, first look for an index.htm/html file. Otherwise, 
	-- serve the directory listing.
	if ( string.endsWith( file, "/" ) ) then
		if (fileExists( file .. "index.htm" )) then 
			file = file .. "index.htm"
		elseif (fileExists( file .. "index.html" )) then
			file = file .. "index.html"
		else
			size = handleDirectory( client, file, path )
			client:close()
			return size, "200"
		end
	end

	-- Open the file and return an error if it fails.
	if ( fileExists( file ) == false ) then 
		client:close()
		return size, "404"
	end;

	-- Find the suffix to get the mime.type.
	_, _, ext  = string.find( file, "\.([^\.]+)$" )
	ext = string.lower(ext)
	if ( ext == nil ) then
		ext = ""
	end
    
	type = mime[ext];
	if ( type == nil ) then type = 'text/plain' end

	f = nil
	o_tries = 0

	f_content = ""

	f = io.open( file, "rb" );

	if not f then return size, "404" end

	f_content = f:read("*a")

	if ext == "lhp" then
		f_content = parse_lhp(file)
	end

	size = string.len(f_content)

	-- Send out the header.
	client:send( "HTTP/1.0 200 OK\r\n" )
	client:send( "Server: psp-httpd " .. appversion .. "\r\n" )
	client:send( "Content-Length: " .. string.format("%u",size)  .. "\r\n" )
	client:send( "Content-Type: " .. type  .. "\r\n" )
	client:send( "Connection: close\r\n\r\n" )
	
	send_data(f_content, client)

	-- Read the file, and then serve it.
	-- I must warn you now that this is still buggy, and there's a chance of
	-- larger binary files being corrupted. I'm working on fixing this.

	if chunkFailures == 1 then
		graphicsPrint("\n\3There was an error sending the file.\n")
	else
		graphicsPrint("\n\4File transfer complete. (" .. friendlySize(size) .. ", " .. c_count .. " chunks of " .. options.chunksize .. "b)\n")
	end

	f:close();

	return size, "200"
end

function send_data(d, client)
	c_count = 0
	bytes_read = 0
	chunkFailures = 0;
	data_length = string.len(d)
	while true do
		bytes = string.sub(d,bytes_read + 1,bytes_read + options.chunksize)
		if bytes ~= nil then
			b_sent = 0;
			iterations = 0;
			chunklength = string.len(bytes)
			while (b_sent < chunklength) and (iterations < options.send_repeats) do
				if Controls.read():start() then break end
				currsent = client:send(string.sub(bytes, b_sent + 1))
				if (currsent > 0) then
					b_sent = b_sent + currsent
				end
				if iterations > 0 then
					System.sleep(50)
				else
					System.sleep(1)
				end
				if iterations == (options.send_repeats - 1) then
					graphicsPrint("\3X")
					chunkFailures = chunkFailures + 1
				end
				iterations = iterations + 1
				graphicsPrint("|")
				totalOutput = totalOutput + currsent
			end
		end
		c_count = c_count + 1
		graphicsPrint("\4.")
		System.sleep(1)
		bytes_read = bytes_read + options.chunksize
		if bytes_read >= data_length then return true end
		if chunkFailures == 1 then return false end
		if Controls.read():start() then return false end
	end
end

dofile "Src/network.lua"
dofile "Src/lhp.lua"

--
-- PROGRAM EXECUTION BEGINS HERE
--

socketCreated = 0
run_s = 1
saveProfile = false
totalOutput = 0
totalInput = 0

while run_s ~= "stop" do

	y = 25
	x = 5

	-- Default MIME types
	mime = { png="image/png", gif="image/gif", jpg="image/jpeg", txt="text/plain", htm="text/html", html="text/html", lhp="text/html", css="text/css", mp3="audio/mpeg3" }

	-- Default options
	options = { wlan_id=0, port=80, chunksize=8192, send_repeats=80, auto_restart=0, autorestart_method=1, root_dir="" }

	-- Load config file and options
	file = io.open("config.txt", "r")
	if file then
		for line in file:lines() do
			equal = string.find(line, "=")
			if equal then
				key = string.sub(line, 1, equal - 1)
				value = string.sub(line, equal + 1)
				if (key ~= "root_dir") then
					value = tonumber(value)
				end
				options[key] = value
			end
		end
		file:close()
	end

	--If the user pressed X at network selection, auto-select connection
	if saveProfile == true then
		state = 1
	else
		state = 0
		selected = 1
		chosen = false
		quit = false
	end

	net_on="f"

	--Select network configuration
	while net_on == "f" do
		net_on = chooseNetwork(options.wlan_id)

		-- If this failed, show the connection selection screen next time, just in case
		if net_on == "f" then
			state = 0
			selected = 1
			chosen = false
			quit = false
			options.wlan_id = 0
		end
	end

	-- Clear screen and buffers
	clear()
	offscreen:clear(Color.new(0, 0, 0))
	
	-- Start the uptime ticker
	uptimeCounter:start()

	-- Start server
	run_s = start_server(options.server_port)

	-- Disconnect from the access point
	Wlan.disconnectAP()
	
	-- Stop the uptime ticker
	uptimeCounter:stop()

end

-- Once the app has terminated, close the WLAN connection
Wlan.term()